"""Declares the data types used in transferring requests to projects."""
from dataclasses import dataclass, field


@dataclass
class Researcher:
    """The dataclass the holds the fields for a Researcher."""
    first_name: str
    last_name: str
    lab_type: str
    email: str
    uri: str


@dataclass
class Project:
    """The dataclass the holds the fields for a Project."""
    name: str
    res: str
    open_date: str = ""
    files: list = field(default_factory=list)
    uri: str = ""


@dataclass
class Container:
    """The dataclass the holds the fields for a Container."""
    name: str = ""
    con_type: str = ""
    con_type_uri: str = ""
    uri: str = ""


@dataclass
class Sample:
    """The dataclass the holds the fields for a Sample."""
    name: str
    udf_to_value: dict = field(default_factory=dict)
    adapter: str = ""
    con: None = field(default=None)
    location: str = "1:1"
    uri: str = ""
    art_uri: str = ""


@dataclass
class CustomForm:
    """The dataclass the holds the fields for a CustomForm."""
    name: str
    req_id: str
    form_id: str
    con_type: str = "Tube"
    field_to_values: dict = field(default_factory=dict)
    samples: list = field(default_factory=list)
    request_type: str = ""


@dataclass
class Service_Price:
    """The dataclass that hold the fields for a services price."""
    price: str
    samples_per_unit: str



#!/usr/bin/env python
"""A host of tools that interact with the ilab REST database."""
from ua_generic_rest_api import ua_generic_rest_api
from bs4 import BeautifulSoup


class IlabApi(ua_generic_rest_api.GenericRestApi):
    """The object that holds the stache environment variables."""
    def __init__(self, core_id, auth_creds):
        host = f"https://api.ilabsolutions.com/v1/cores/{core_id}/"
        super().__init__(host, auth_creds, "page")

    def get(self, endpoints, parameters=None, get_all=True, total_pages=None):
        response = super().get(endpoints, parameters)
        if get_all:
            first_page_soup = BeautifulSoup(response[0].text, "xml")
            total_pages = first_page_soup.find("total-pages")
            if total_pages:
                total_pages = int(total_pages.text)
                return super().get(endpoints, parameters, total_pages)

        return response



"""Tools that interact with Ilab's REST database."""
import re
import copy
import traceback
from bs4 import BeautifulSoup
from ua_ilab_tools import extract_custom_forms, ilab_api, api_types

ONLY_INT_FIELDS = [
    "Concentration_each_sample", "Concentration", "Volume (uL)",
    "Initial_Number_Slides_or_Punches_each_sample", "Template Length",
    "Template_Length_each_sample"]


SKIP_FORM_PATTERNS = [r"REQUEST A QUOTE.*", r".*NQ.*"]


class IlabConfigError(Exception):
    """The request or form has been configured or altered incorrectly."""


class IlabTools():
    def __init__(self, core_id, token):
        if "Bearer" not in token:
            token = "Bearer " + token
        auth_creds = {
            "Authorization": f"{token}",
            "Content-Type": "application/xml"
        }
        self.api = ilab_api.IlabApi(core_id, auth_creds)

    def get_service_requests(
        self,
        status="processing",
        from_date="2015-01-01T12:00Z",
        specific_uri=None
    ):
        """Get the service requests with the given status from ilab's REST DB.

        Keyword Arguments:
            status (string):
                If you want service requests with a certain status. By default,
                it is 'processing'.
            specific_uri (string):
                If you want a specific endpoint.

        Returns:
            req_uri_to_soup (dict):
                The dictionary that holds all of the
                {service req uris:request soup}. If no service requests are
                found, returns an empty dict.
        """
        req_uri_to_soup = {}
        if specific_uri:
            get_responses = self.api.get(
                f"service_requests/{specific_uri}.xml", get_all=False)
            requests_soup = BeautifulSoup(get_responses[0].text, "xml")
            requests_soup = requests_soup.find("service-request")
            req_uri_to_soup[requests_soup.find("id").string] = requests_soup
        else:
            get_responses = self.api.get(
                "service_requests.xml",
                parameters={"states": status, "from_date": from_date},
                get_all=True)

        # Soup all get responses (multiple pages or not).
        req_paged_soups = [
            BeautifulSoup(response.text, "xml") for response in get_responses]

        # Get every service-request in every page.
        for get_soup in req_paged_soups:
            for req_soup in get_soup.find_all("service-request"):
                req_uri_to_soup[req_soup.find("id").string] = req_soup

        return req_uri_to_soup

    def get_service_cost(self, price_id):
        """Get the cost associated with the given service_id.

        Arguments:
            price_id (string):
                The id associated with a price.

        Returns:
            service_price (namedtuple):
                The calculated price of the service, or None if not found.
        """
        service_price = None
        get_responses = self.api.get("services.xml")

        for response in get_responses:
            services_page_soup = BeautifulSoup(response.text, "xml")
            id_soup = services_page_soup.find(string=price_id)
            if id_soup:
                service_soup = id_soup.find_parent("service")
                price_soup = service_soup.find("price")
                current_price = price_soup.find("price").string
                unit = price_soup.find("unit").find("description").string
                service_price = api_types.Service_Price(
                    price=float(current_price),
                    samples_per_unit=unit)

        return service_price

    def get_request_charges(self, req_id):
        """Get all of the charges of the req_id passed in.

        Arguments:
            req_id(string):
                The unique string of ints that map to a request.

        Returns:
            charges_uri_soup (dict):
                The dict of uri_to_soup of all the charges associated with that
                request. Returns an empty dict if not found.
        """
        get_responses = self.api.get(f"service_requests/{req_id}/charges.xml")
        charge_paged_soups = [
            BeautifulSoup(response.text, "xml") for response in get_responses]

        charges_uri_soup = dict()
        for get_soup in charge_paged_soups:
            for charge in get_soup.find_all("charge"):
                charges_uri_soup[charge.find("id").string] = charge

        return charges_uri_soup

    def get_milestones(self, request_id):
        """Get all of the milestones associated with a service request.

        Arguments:
            request_id (string):
                The unique string of ints that map to a request.

        Returns:
            milestone_name_soup (dict):
                Holds all {milestone name : soup of milestone}. Returns an
                empty dict if not found.
        """
        get_responses = self.api.get(
            f"service_requests/{request_id}/milestones.xml")
        milestone_paged_soups = [
            BeautifulSoup(response.text, "xml") for response in get_responses]

        milestone_name_soup = {}
        for get_soup in milestone_paged_soups:
            for milestone in get_soup.find_all("milestone"):
                name_tag = milestone.find("name")
                if name_tag:
                    milestone_name_soup[name_tag.string] = milestone

        return milestone_name_soup

    def get_custom_forms(self, req_id):
        """Get all of the custom forms of the req_id passed in.

        Arguments:
            req_id (string):
                The unique string of ints that map to a request.

        Returns:
            forms_uri_to_soup (dict):
                The dictionary that holds all of the
                {custom form uris: form_soup}. Returns an empty dict if not
                found.
        """
        get_responses = self.api.get(
            f"service_requests/{req_id}/custom_forms.xml")
        form_paged_soups = [
            BeautifulSoup(response.text, "xml") for response in get_responses]

        forms_uri_to_soup = {}
        for get_soup in form_paged_soups:
            for form in get_soup.find_all("custom-form"):
                forms_uri_to_soup[form.find("id").string] = form

        return forms_uri_to_soup


def extract_project_info(req_soup, full_name=False):
    """Extract the relevant project info from a request.

    Arguments:
        req_soup (BS4 soup object):
            The soup of the request.
        full_name (boolean):
            Whether or not to capture the entire project name or just the last
            hyphenated element.

    Returns:
        prj_info (Project):
            The required info to post a project.
    """
    if full_name:
        prj_name = req_soup.find("name").string
    else:
        prj_name = req_soup.find("name").string.split('-')[-1]
    res_name = req_soup.find("owner").find("name").string
    email = req_soup.find("owner").find("email").string
    # NOTE: Change this line to your own institution's email domain.
    if "email.arizona.edu" in email:
        res_lab = "internal"
    else:
        res_lab = "external"

    # Replace all not ascii chars with ascii ones, and any symbols with '-'.
    prj_res = api_types.Researcher(
        extract_custom_forms._sanitize_text(res_name.split()[0]),
        extract_custom_forms._sanitize_text(res_name.split()[-1]),
        extract_custom_forms._sanitize_text(res_lab),
        email,
        "")
    prj_info = api_types.Project(prj_name, prj_res)

    return prj_info


def extract_custom_form_info(req_id, form_id, form_soup):
    """Extract all of the fields passed into the form.

    Arguments:
        req_id (String):
            The unique string of ints that map to a request (URI).
        form_id (String):
            The unique string of ints that map to a form.
        form_soup (BeautifulSoup object):
            The soup of the form you want to parse.

    Returns:
        form_info (CustomForm):
            The CustomForm object with all of the form's fields initialized.

    Raises:
        TypeError:
            The form has no fields configured.
        ValueError:
            The form has duplicate samples.
    """

    # If we need any of these types, we can make new methods.
    skip_types = ["charges", "file", "table", "help", "file_no_upload"]
    field_strategy = {
        "handsontable_grid": extract_custom_forms.grid_type,
        "checkbox": extract_custom_forms.checkbox_type,
        "all_others": extract_custom_forms.all_other_types}

    # Find the desired custom form out of all of the form_soup.
    target_form = form_soup.find(string=form_id)
    target_form = target_form.find_parent("custom-form")
    form_soup = target_form

    form_name = form_soup.find("name").string
    fields_soup = form_soup.find("fields")
    form_info = api_types.CustomForm(form_name, req_id, form_id)

    # Get all of the field information.
    for field_soup in fields_soup.find_all("field"):
        field_type = field_soup.find("type").string
        if field_type in skip_types:
            # Do nothing with the field types that we don't yet care about.
            continue

        try:
            field_strategy[field_type](field_soup, form_info)
        except KeyError:
            field_strategy["all_others"](field_soup, form_info)
        except TypeError:
            raise TypeError(
                f"The grid in the {form_info.name} form in request"
                f" {form_info.req_id} has been filled out incorrectly. The"
                f" error message is: {traceback.format_exc()}")

    # Raise an error if a form doesn't have samples.
    if not form_info.samples:
        return form_info

    if form_info.field_to_values.get("duplicate_samples"):
        if form_info.field_to_values["duplicate_samples"] == "Yes":
            b_samples = copy.deepcopy(form_info.samples)
            for a_sample, b_sample in zip(form_info.samples, b_samples):
                a_sample.name += "A"
                b_sample.name += "B"
            form_info.samples = form_info.samples + b_samples

    extract_custom_forms.bind_container_info(form_info)

    # Allows duplicate names if they have different well locations in a
    # plate.
    if form_info.con_type != "96 well plate":
        sample_names = [sample.name for sample in form_info.samples]
        if len(set(sample_names)) != len(sample_names):
            raise ValueError(
                f"There are two or more samples named the same thing in"
                f" request {form_info.req_id}. Please review and edit your"
                f" sample names.")

    for name, value in form_info.field_to_values.items():
        if name in ONLY_INT_FIELDS:
            value = re.sub(r"[^.0-9]", "", value)
        if "_each_sample" in name:
            udf_name = name.replace("_each_sample", "").replace("_", " ")
            for sample in form_info.samples:
                sample.udf_to_value[udf_name] = value

    return form_info



"""Methods to harvest data from iLab custom forms."""
import re
import math
import unicodedata
import ast
from ua_ilab_tools import api_types


# NOTE: Change these globals to match names of these types of fields in your
# Clarity instance.


DISALLOWED_CHARS = r""
REPLACE_CHARS = r""


ONLY_INT_FIELDS = [
    "Concentration_each_sample", "Concentration", "Volume (uL)",
    "Initial_Number_Slides_or_Punches_each_sample", "Template Length",
    "Template_Length_each_sample"]


BOOL_FIELDS = [
    "H&E Slide Submitted?"]


def update_globals(disallowed_chars, replace_chars):
    global DISALLOWED_CHARS
    DISALLOWED_CHARS = disallowed_chars
    global REPLACE_CHARS
    REPLACE_CHARS = replace_chars


def grid_type(field_soup, form_info):
    """Harvest all of the sample info and add it to the CustomForm.

    Arguments:
        field_soup (BeautifulSoup object): The soup containing the
            field information.
        form_info (api_types.CustomForm): the dataclass holding info on the
            form.
    """
    if field_soup.find("value") is None:
        return
    elif field_soup.find("value").string == "[]":
        # The list is empty, do nothing.
        return

    if form_info.samples:
        # Sample data are in 2 or more grids.
        raise TypeError(
            f"There are 2 or more grids with sample data in the"
            f" {form_info.name} Custom Form for Request:"
            f" {form_info.req_id} . Please delete the data from the grid"
            f" that you don't want to use and try again.")
    form_info.samples.extend(extract_from_grid(field_soup, form_info))


def checkbox_type(field_soup, form_info):
    """Gather info from the ilab_form checkbox type.

    Arguments:
        field_soup (BeautifulSoup object): The soup containing the
            field information.
        form_info (api_types.CustomForm): the dataclass holding info on the
            form.
    """
    field_name = field_soup.find("identifier").string
    if field_soup.find("value"):
        form_info.field_to_values[field_name] = True
    else:
        form_info.field_to_values[field_name] = False


def all_other_types(field_soup, form_info):
    """All other types of field where data is saved as a form field.

    These types include: small (string), medium (text_medium),
        large (text) text box, date, select (pull down menu), radio,
        and text_section fields.

    Arguments:
        field_soup (BeautifulSoup object): The soup containing the
            field information.
    """
    value = field_soup.find("value")
    if value and value.string:
        form_info.field_to_values[field_soup.find(
            "identifier").string] = field_soup.find("value").string

    if value == "container_name":
        form_info.con_type = "96 well plate"


def bind_container_info(form_info):
    """Associate the container info to the form samples.

    Arguments:
        form_info: The form that holds the samples that are being bound.

    Restrictions:
        One container type per samples list.
    """
    con_strategy = {
        "Tube": tube_bind,
        "8-well strip": eight_well_strip_bind,
        "96 well plate": plate_96_bind}

    # Every sample assuredly has the same con_type.
    con_strategy[form_info.con_type](form_info)


def tube_bind(form_info):
    """Add the tube container info to the samples in the form.

    Arguments:
        form_info (api_types.CustomForm): the dataclass holding info on the
            form.
    """
    sample_names = []
    for sample in form_info.samples:
        sample.con = api_types.Container(sample.name, "Tube")
        sample.con.name = sample.name
        sample_names.append(sample.name)


def plate_96_bind(form_info):
    """Add the plate container info to the samples in the form.

    Arguments:
        form_info (api_types.CustomForm): the dataclass holding info on the
            form.
    """

    smp_locations = {}
    for sample in form_info.samples:
        sample.con = api_types.Container()
        # If 'Container Name' in grid.
        if sample.udf_to_value.get("Container Name"):
            con_name = sample.udf_to_value["Container Name"]
        # If 'container_name' in form.
        elif form_info.field_to_values.get("container_name"):
            con_name = form_info.field_to_values["container_name"]
        else:
            raise TypeError(
                f"The sample {sample.name} in request {form_info.req_id} on"
                f" form {form_info.name} was not given a container name.")

        # Replace all not ascii chars with ascii ones.
        con_name = _sanitize_text(con_name)
        con_name = re.sub(r"[^a-zA-Z0-9\-]", "-", con_name)
        sample.con.name = con_name
        sample.con.con_type = "96 well plate"

        # Make a dictionary of con_name: well, to make sure that there are no
        # two samples in the same well.
        smp_locations.setdefault(
            sample.con.name, []).append(sample.location)
    for locs in smp_locations.values():
        if len(set(locs)) != len(locs):
            raise TypeError(
                f"There are two or more samples with the same well"
                f" location in request {form_info.req_id}. Please review and"
                f" edit your well locations.")


def eight_well_strip_bind(form_info):
    """Add the eight-well-strip container info to the samples in the form.

    Arguments:
        form_info (api_types.CustomForm): the dataclass holding info on the
            form.
    """
    container_names = []
    tube_numbers = [sample.location for sample in form_info.samples]
    con_num = math.ceil(len(form_info.samples) / 8)

    if len(set(tube_numbers)) != len(tube_numbers):
        raise TypeError(
            f"There are two or more samples with the same tube number in"
            f" request {form_info.req_id}. Please review and edit your"
            f" tube numbers.")

    # Create the list of names for each strip submitted.
    for num in range(con_num):
        container_names.append(f"{form_info.req_id}-{num + 1}")
    for num, sample in enumerate(form_info.samples):
        sample.con = api_types.Container()
        sample.con.name = container_names[num // 8]
        sample.con.con_type = "8-well strip"


def extract_from_grid(field_soup, form_info):
    """Return list of Samples, extracted from grid portion of form xml.

    Arguments:
        field_soup (BeautifulSoup object): The soup of the grid field.
        form_info (api_types.CustomForm): dataclass holding information for the
            form.

    Returns:
        samples (list of Samples): The samples extracted from the grid with as
            much information that the grid provided.

    Restrictions:
        Every grid must start with the sample name (though the header
            doesn't matter).
        The Well Location or Tube Number cannot be empty if the column is
            in the grid or an error will be thrown.
        To allow a submission to have more than one plate on it, place the
            column name "Container Name" in the grid.
        To throw no errors, the columns that hold information for these
            data must be named:
                well location info = "Well Location".
                tube number for 8-strip tubes = "Tube Number".
                container name if the grid holds info for > 1
                    container = "Container Name".
                concentration info = "Concentration (ng/uL)" or
                    "Average Template Concentration (ng/uL)".
                TGM assay names = "TGM Assay".
                Sanger primers = "LVS Primer" or "HVS Primer".
    """

    samples = list()
    column_names = field_soup.find("default").string.split(',')
    sample_values = field_soup.find("value").string
    sample_values = sample_values.replace("null", "None")

    for value in ast.literal_eval(sample_values):
        # If there is not a real udf value, skip it.
        if value[0] in ["", None]:
            continue

        # remove whitespace.
        new_sample_name = value[0].strip()
        # Sanitize the name of the sample.
        new_sample_name = _sanitize_text(new_sample_name)
        # Remove '[:,]' from sample names and replace it with '-'.
        new_sample_name = re.sub(r"[:,.]", "-", new_sample_name)
        new_sample = api_types.Sample(new_sample_name)

        # skip "empty" sample names.
        if new_sample.name.upper().strip() == "EMPTY":
            continue

        # Skip the first column, as that must always be the sample name.
        for udf_name, udf_value in zip(column_names[1:], value[1:]):
            new_sample = udf_parser(
                form_info, new_sample, udf_name, udf_value)

        samples.append(new_sample)

    for sample in samples:
        if "TGM Assay List" in sample.udf_to_value.keys():
            sample.udf_to_value["TGM Assay List"] = sample.udf_to_value[
                "TGM Assay List"].strip(',')

        if "Primer" in sample.udf_to_value.keys():
            sample.udf_to_value["Primer"] = sample.udf_to_value[
                "Primer"].strip(',')

    return samples


def udf_parser(form, new_sample, udf_name, udf_value):
    """Parses udf data from form grid and returns the udf-updated sample.

    Arguments:
        form (api_types.CustomForm): dataclass holding information for the
            form.
        new_sample (api_types.Sample): a Sample dataclass to update with udf
            info.
        udf_name (string): the name of the udf we want to update new_sample
            with.
        udf_value (string): The value to add to new_sample.

    Returns:
        Updated new_sample.
    """
    # Check if there is a poorly inputted sample.
    udf_name = udf_name.strip()
    if udf_value in ["", None]:
        if udf_name in ["Well Location", "Tube Number"]:
            raise TypeError(
                f"The sample {new_sample.name} was given no well location.")
        # Don't add empty values as udf's, and return the sample and
        # con_type unchanged.
        return new_sample

    # Sanitize the udf_value.
    udf_value = _sanitize_text(udf_value)
    # Remove whitespace.
    udf_value = udf_value.strip()

    # Form specific handling.
    # Extract well info for a plate.
    if udf_name == "Well Location":
        new_sample.location = _translate_to_well(udf_value)
        form.con_type = "96 well plate"
        # Do not add Well Location as a udf, and return the modified sample.
        return new_sample

    elif udf_name == "Container Name":
        new_sample.udf_to_value["Container Name"] = re.sub(
            r"[:,.]", "-", udf_value)
        form.con_type = "96 well plate"
        return new_sample

    # Extract well info for strip tubes.
    elif udf_name in ["Tube Number", "Well Number (1-8)"]:
        udf_value = udf_value.upper()
        form.con_type = "8-well strip"
        new_sample.location = _translate_to_well(udf_value, plate=False)
        # Do not add Tube Number as a udf, and return the modified sample.
        return new_sample

    # Adds the adapter information to it's own field.
    elif "Adapter" in udf_name:
        new_sample.adapter = udf_value
        return new_sample

    elif udf_name == "Average Template Length":
        udf_name = "Template Length"

    elif (udf_name in [
            "Average Template Concentration (ng/uL)",
            "Concentration (ng/uL)"]):
        udf_name = "Concentration"

    elif "TGM Assay" in udf_name:
        udf_name = "TGM Assay List"
        new_sample.udf_to_value.setdefault(udf_name, "")
        udf_value = new_sample.udf_to_value[udf_name] + f"{udf_value},"

    elif "HVS Primer" in udf_name or "LVS Primer" in udf_name:
        udf_name = "Primer"
        new_sample.udf_to_value.setdefault("Primer", "")
        udf_value = new_sample.udf_to_value["Primer"] + f"{udf_value},"

    # Scrub client-inputted data that must be numbers AFTER correcting the
    # udf_names (e.g. if the form calls Concentration and Template Length
    # different things, the value still needs to be scrubbed).
    if udf_name in ONLY_INT_FIELDS:
        udf_value = re.sub(r"[^0-9.]", "", udf_value)

    if udf_name in BOOL_FIELDS:
        if udf_value.upper() in ["YES", "TRUE", "AFFIRMATIVE", "Y", "T"]:
            udf_value = "true"
        else:
            udf_value = "false"

    new_sample.udf_to_value[udf_name] = udf_value

    return new_sample


def _sanitize_text(text):
    """Convert text to ascii, replace chars [^a-zA-Z0-9:,.] with '-'."""
    # Replace all not ascii chars with ascii ones
    text = unicodedata.normalize("NFKD", text).encode(
        "ascii", "ignore").decode("ascii")
    # Convert anything that is not alphanumeric or a hyphen to a hyphen.
    text = re.sub(DISALLOWED_CHARS, REPLACE_CHARS, text)
    # Convert '+' to 'plus' to avoid the confusion of changing a '+' to
    # a '-'.
    text = re.sub(r"[+]", "plus", text)

    return text


def _translate_to_well(well, plate=True):
    """Translates any allowed string options A1, A01, 1, or A:01 to A:1."""
    well = well.upper().strip()
    # Do nothing if A:1 - H:12.
    if re.search(r"[A-H]:[1-9][0-3]{0,1}", well) is None:
        # A1-H12.
        if re.search(r"[A-H][1-9][0-3]{0,1}", well):
            well = well[0] + ':' + well[1:]
        # A01-H09 or A:01 - H:09.
        elif re.search(r"[A-H]:{0,1}0[1-9]", well):
            if ':' not in well:
                well = well.replace('0', ':')
            else:
                well = well.replace('0', '')
        # If the well is 1-8, treat as a strip tube.
        elif re.search(r"^[1-8]$", well):
            # If someone has inputted numbers instead of well locations
            #  for a plate format, raise an error.
            if plate:
                raise TypeError(
                    "We don't accept plate indexes of that type.")
            # Convert 1-8 into the strip tube format.
            well = f"A:{well}"
        # Anything else.
        else:
            raise TypeError("We don't accept indexes of that type.")

    return well



import os
from setuptools import setup, find_packages


def readme(filename):
    full_path = os.path.join(os.path.dirname(__file__), filename)
    with open(full_path, 'r') as file:
        return file.read()


setup(
    name="ua_ilab_tools",
    version="2.1.1",
    packages=find_packages(),
    author="Stephen Stern, Rafael Lopez, Etienne Thompson",
    author_email="sterns1@email.arizona.edu",
    include_package_data=True,
    long_description=readme("README.md"),
    long_description_content_type='text/markdown',
    url="https://github.com/UACoreFacilitiesIT/UA-Ilab-Tools",
    license="MIT",
    description=(
        "Tools that interact with Agilent's iLab REST architecture."),
    install_requires=[
        "bs4",
        "ua-generic-rest-api",
    ],
)
